#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

import argparse
import sys
from pathlib import Path

import qq
import pkg_resources
import aiohttp
import platform


def show_version():
    entries = []

    entries.append('- Python v{0.major}.{0.minor}.{0.micro}-{0.releaselevel}'.format(sys.version_info))
    version_info = qq.version_info
    entries.append('- qq.py v{0.major}.{0.minor}.{0.micro}-{0.releaselevel}'.format(version_info))
    if version_info.releaselevel != 'final':
        pkg = pkg_resources.get_distribution('qq.py')
        if pkg:
            entries.append(f'    - qq.py pkg_resources: v{pkg.version}')

    entries.append(f'- aiohttp v{aiohttp.__version__}')
    uname = platform.uname()
    entries.append('- system info: {0.system} {0.release} {0.version}'.format(uname))
    print('\n'.join(entries))


def core(parser, args):
    if args.version:
        show_version()


_bot_template = """#!/usr/bin/env python3
from qq.ext import commands
import qq
import config
class Bot(commands.{base}):
    def __init__(self, **kwargs):
        super().__init__(command_prefix=commands.when_mentioned_or('{prefix}'), **kwargs)
        for cog in config.cogs:
            try:
                self.load_extension(cog)
            except Exception as exc:
                print(f'Could not load extension {{cog}} due to {{exc.__class__.__name__}}: {{exc}}')
    async def on_ready(self):
        print(f'Logged on as {{self.user}} (ID: {{self.user.id}})')
bot = Bot()
# write general commands here
bot.run(config.token)
"""

_gitignore_template = """# Byte-compiled / optimized / DLL files
__pycache__/
*.py[cod]
*$py.class
# C extensions
*.so
# Distribution / packaging
.Python
env/
build/
develop-eggs/
dist/
downloads/
eggs/
.eggs/
lib/
lib64/
parts/
sdist/
var/
*.egg-info/
.installed.cfg
*.egg
# Our configuration files
config.py
"""

_cog_template = '''from qq.ext import commands
import qq
class {name}(commands.Cog{attrs}):
    """The description for {name} goes here."""
    def __init__(self, bot):
        self.bot = bot
{extra}
def setup(bot):
    bot.add_cog({name}(bot))
'''

_cog_extras = '''
    def cog_unload(self):
        # clean up logic goes here
        pass
    async def cog_check(self, ctx):
        # checks that apply to every command in here
        return True
    async def bot_check(self, ctx):
        # checks that apply to every command to the bot
        return True
    async def bot_check_once(self, ctx):
        # check that apply to every command but is guaranteed to be called only once
        return True
    async def cog_command_error(self, ctx, error):
        # error handling to every command in here
        pass
    async def cog_before_invoke(self, ctx):
        # called before a command is called here
        pass
    async def cog_after_invoke(self, ctx):
        # called after a command is called here
        pass
'''

# certain file names and directory names are forbidden
# see: https://msdn.microsoft.com/en-us/library/windows/desktop/aa365247%28v=vs.85%29.aspx
# although some of this doesn't apply to Linux, we might as well be consistent
_base_table = {
    '<': '-',
    '>': '-',
    ':': '-',
    '"': '-',
    # '/': '-', these are fine
    # '\\': '-',
    '|': '-',
    '?': '-',
    '*': '-',
}

# NUL (0) and 1-31 are disallowed
_base_table.update((chr(i), None) for i in range(32))

_translation_table = str.maketrans(_base_table)


def to_path(parser, name, *, replace_spaces=False):
    if isinstance(name, Path):
        return name

    if sys.platform == 'win32':
        forbidden = ('CON', 'PRN', 'AUX', 'NUL', 'COM1', 'COM2', 'COM3', 'COM4', 'COM5', 'COM6', 'COM7', \
                     'COM8', 'COM9', 'LPT1', 'LPT2', 'LPT3', 'LPT4', 'LPT5', 'LPT6', 'LPT7', 'LPT8', 'LPT9')
        if len(name) <= 4 and name.upper() in forbidden:
            parser.error('invalid directory name given, use a different one')

    name = name.translate(_translation_table)
    if replace_spaces:
        name = name.replace(' ', '-')
    return Path(name)


def newbot(parser, args):
    new_directory = to_path(parser, args.directory) / to_path(parser, args.name)

    # as a note exist_ok for Path is a 3.5+ only feature
    # since we already checked above that we're >3.5
    try:
        new_directory.mkdir(exist_ok=True, parents=True)
    except OSError as exc:
        parser.error(f'could not create our bot directory ({exc})')

    cogs = new_directory / 'cogs'

    try:
        cogs.mkdir(exist_ok=True)
        init = cogs / '__init__.py'
        init.touch()
    except OSError as exc:
        print(f'warning: could not create cogs directory ({exc})')

    try:
        with open(str(new_directory / 'config.py'), 'w', encoding='utf-8') as fp:
            fp.write('token = "place your token here"\ncogs = []\n')
    except OSError as exc:
        parser.error(f'could not create config file ({exc})')

    try:
        with open(str(new_directory / 'bot.py'), 'w', encoding='utf-8') as fp:
            base = 'Bot' if not args.sharded else 'AutoShardedBot'
            fp.write(_bot_template.format(base=base, prefix=args.prefix))
    except OSError as exc:
        parser.error(f'could not create bot file ({exc})')

    if not args.no_git:
        try:
            with open(str(new_directory / '.gitignore'), 'w', encoding='utf-8') as fp:
                fp.write(_gitignore_template)
        except OSError as exc:
            print(f'warning: could not create .gitignore file ({exc})')

    print('successfully made bot at', new_directory)


def newcog(parser, args):
    cog_dir = to_path(parser, args.directory)
    try:
        cog_dir.mkdir(exist_ok=True)
    except OSError as exc:
        print(f'warning: could not create cogs directory ({exc})')

    directory = cog_dir / to_path(parser, args.name)
    directory = directory.with_suffix('.py')
    try:
        with open(str(directory), 'w', encoding='utf-8') as fp:
            attrs = ''
            extra = _cog_extras if args.full else ''
            if args.class_name:
                name = args.class_name
            else:
                name = str(directory.stem)
                if '-' in name or '_' in name:
                    translation = str.maketrans('-_', '  ')
                    name = name.translate(translation).title().replace(' ', '')
                else:
                    name = name.title()

            if args.display_name:
                attrs += f', name="{args.display_name}"'
            if args.hide_commands:
                attrs += ', command_attrs=dict(hidden=True)'
            fp.write(_cog_template.format(name=name, extra=extra, attrs=attrs))
    except OSError as exc:
        parser.error(f'could not create cog file ({exc})')
    else:
        print('successfully made cog at', directory)


def add_newbot_args(subparser):
    parser = subparser.add_parser('newbot', help='快速创建命令机器人项目')
    parser.set_defaults(func=newbot)

    parser.add_argument('name', help='机器人项目名称')
    parser.add_argument('directory', help='放置它的目录（默认：.）', nargs='?', default=Path.cwd())
    parser.add_argument('--prefix', help='机器人前缀（默认：$）', default='$', metavar='<prefix>')
    parser.add_argument('--sharded', help='是否使用 AutoShardedBot', action='store_true')
    parser.add_argument('--no-git', help='不要创建 .gitignore 文件', action='store_true', dest='no_git')


def add_newcog_args(subparser):
    parser = subparser.add_parser('newcog', help='快速创建一个新的 cog 模板')
    parser.set_defaults(func=newcog)

    parser.add_argument('name', help='齿轮名称')
    parser.add_argument('directory', help='放置它的目录（默认：cogs）', nargs='?',
                        default=Path('cogs'))
    parser.add_argument('--class-name', help='cog 的类名（默认：<name>）', dest='class_name')
    parser.add_argument('--display-name', help='齿轮名称（默认值：<name>）')
    parser.add_argument('--hide-commands', help='是否隐藏 cog 中的所有命令', action='store_true')
    parser.add_argument('--full', help='添加所有特殊方法', action='store_true')


def parse_args():
    parser = argparse.ArgumentParser(prog='qq', description='qq.py 帮助工具')
    parser.add_argument('-v', '--version', action='store_true', help='显示库版本')
    parser.set_defaults(func=core)

    subparser = parser.add_subparsers(dest='subcommand', title='subcommands')
    add_newbot_args(subparser)
    add_newcog_args(subparser)
    return parser, parser.parse_args()


def main():
    parser, args = parse_args()
    args.func(parser, args)


if __name__ == '__main__':
    main()



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from typing import List, Literal, Optional, TypedDict, Union

from .embed import Embed
from .emoji import PartialEmoji
from .member import Member
from .user import User
from ..utils import SnowflakeList


class _AttachmentOptional(TypedDict, total=False):
    height: Optional[int]
    width: Optional[int]
    content_type: str


class Attachment(_AttachmentOptional):
    id: int
    filename: str
    size: int
    url: str


class _MessageOptional(TypedDict, total=False):
    guild_id: str


class Message(_MessageOptional):
    id: str
    channel_id: str
    author: User
    member: Member
    content: str
    timestamp: str
    edited_timestamp: Optional[str]
    mention_everyone: bool
    mentions: List[User]
    attachments: List[Attachment]
    embeds: List[Embed]
    reactions: List[Reaction]


AllowedMentionType = Literal['roles', 'users', 'everyone']


class AllowedMentions(TypedDict):
    parse: List[AllowedMentionType]
    roles: SnowflakeList
    users: SnowflakeList
    replied_user: bool


class MessageReference(TypedDict, total=False):
    message_id: int
    channel_id: int
    guild_id: int
    fail_if_not_exists: bool


class Reaction(TypedDict):
    count: int
    me: bool
    emoji: PartialEmoji



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import asyncio
import collections
import collections.abc
import inspect
import importlib.util
import sys
import traceback
import types
from typing import Any, Callable, Mapping, List, Dict, TYPE_CHECKING, Optional, TypeVar, Type, Union

import qq

from .core import GroupMixin
from .view import StringView
from .context import Context
from . import errors
from .help import HelpCommand, DefaultHelpCommand
from .cog import Cog

if TYPE_CHECKING:
    import importlib.machinery

    from qq.message import Message
    from ._types import (
        Check,
        CoroFunc,
    )

__all__ = (
    'when_mentioned',
    'when_mentioned_or',
    'Bot',
    'AutoShardedBot',
)

MISSING: Any = qq.utils.MISSING

T = TypeVar('T')
CFT = TypeVar('CFT', bound='CoroFunc')
CXT = TypeVar('CXT', bound='Context')


def when_mentioned(bot: Union[Bot, AutoShardedBot], msg: Message) -> List[str]:
    """实现与提及的命令前缀等效的可调用对象。

    这些旨在传递到 :attr:`.Bot.command_prefix` 属性。
    """
    # bot.user will never be None when this is called
    return [f'<@{bot.user.id}> ', f'<@!{bot.user.id}> ']  # type: ignore


def when_mentioned_or(*prefixes: str) -> Callable[[Union[Bot, AutoShardedBot], Message], List[str]]:
    """在提及或提供其他前缀时实现的可调用对象。

    这些旨在传递到 :attr:`.Bot.command_prefix` 属性。

    Example
    --------

    .. code-block:: python3

        bot = commands.Bot(command_prefix=commands.when_mentioned_or('!'))


    .. note::

        此可调用对象返回另一个可调用对象，因此如果这是在自定义可调用对象中完成的，则必须调用返回的可调用对象，例如：

        .. code-block:: python3

            async def get_prefix(bot, message):
                extras = await prefixes_for(message.guild) # returns a list
                return commands.when_mentioned_or(*extras)(bot, message)


    See Also
    ----------
    :func:`.when_mentioned`
    """

    def inner(bot, msg):
        r = list(prefixes)
        r = when_mentioned(bot, msg) + r
        return r

    return inner


def _is_submodule(parent: str, child: str) -> bool:
    return parent == child or child.startswith(parent + ".")


class _DefaultRepr:
    def __repr__(self):
        return '<default-help-command>'


_default = _DefaultRepr()


class BotBase(GroupMixin):
    def __init__(self, command_prefix, help_command=_default, description=None, **options):
        super().__init__(**options)
        self.command_prefix = command_prefix
        self.extra_events: Dict[str, List[CoroFunc]] = {}
        self.__cogs: Dict[str, Cog] = {}
        self.__extensions: Dict[str, types.ModuleType] = {}
        self._checks: List[Check] = []
        self._check_once = []
        self._before_invoke = None
        self._after_invoke = None
        self._help_command = None
        self.description = inspect.cleandoc(description) if description else ''
        self.owner_id = options.get('owner_id')
        self.owner_ids = options.get('owner_ids', set())
        self.strip_after_prefix = options.get('strip_after_prefix', False)

        if self.owner_id and self.owner_ids:
            raise TypeError('owner_id 和 owner_ids 都被设置。')

        if not self.owner_id and not self.owner_ids:
            raise TypeError('owner_id 和 owner_ids 必须设置其中一个。')

        if self.owner_ids and not isinstance(self.owner_ids, collections.abc.Collection):
            raise TypeError(f'owner_ids 必须是一个集合，而不是 {self.owner_ids.__class__!r}')

        if help_command is _default:
            self.help_command = DefaultHelpCommand()
        else:
            self.help_command = help_command

    # internal helpers

    def dispatch(self, event_name: str, *args: Any, **kwargs: Any) -> None:
        # super() will resolve to Client
        super().dispatch(event_name, *args, **kwargs)  # type: ignore
        ev = 'on_' + event_name
        for event in self.extra_events.get(ev, []):
            self._schedule_event(event, ev, *args, **kwargs)  # type: ignore

    @qq.utils.copy_doc(qq.Client.close)
    async def close(self) -> None:
        for extension in tuple(self.__extensions):
            try:
                self.unload_extension(extension)
            except Exception:
                pass

        for cog in tuple(self.__cogs):
            try:
                self.remove_cog(cog)
            except Exception:
                pass

        await super().close()  # type: ignore

    async def on_command_error(self, context: Context, exception: errors.CommandError) -> None:
        """|coro|

        机器人提供的默认命令错误处理程序。

        默认情况下，这会打印到 :data:`sys.stderr` 但是它可以被覆盖以使用不同的实现。

        仅当你没有为命令错误指定任何监听器时才会触发。
        """
        if self.extra_events.get('on_command_error', None):
            return

        command = context.command
        if command and command.has_error_handler():
            return

        cog = context.cog
        if cog and cog.has_error_handler():
            return

        print(f'忽略命令 {context.command} 中的异常：', file=sys.stderr)
        traceback.print_exception(type(exception), exception, exception.__traceback__, file=sys.stderr)

    # global check registration

    def check(self, func: T) -> T:
        r"""向机器人添加全局检查的装饰器。

        全局检查类似于基于每个命令应用的 :func:`.check`，不同之处在于它在任何命令检查被验证之前运行并适用于机器人拥有的每个命令。

        .. note::

            此函数可以是常规函数或协程。

        类似于命令 :func:`.check` ，它接受一个类型为 :class:`.Context` 的参数，并且只能引发从 :exc:`.CommandError` 继承的异常。

        Example
        ---------

        .. code-block:: python3

            @bot.check
            def check_commands(ctx):
                return ctx.command.qualified_name in allowed_commands

        """
        # T was used instead of Check to ensure the type matches on return
        self.add_check(func)  # type: ignore
        return func

    def add_check(self, func: Check, *, call_once: bool = False) -> None:
        """向机器人添加全局检查。

        这是 :meth:`.check` 和 :meth:`.check_once` 的非装饰器接口。

        Parameters
        -----------
        func
            用作全局检查的函数。
        call_once: :class:`bool`
            如果为 ``True`` 每个 :meth:`.invoke` 调用只应调用一次该函数。
        """

        if call_once:
            self._check_once.append(func)
        else:
            self._checks.append(func)

    def remove_check(self, func: Check, *, call_once: bool = False) -> None:
        """从机器人中删除全局检查。

        此函数是幂等的，如果该函数不在全局检查中，则不会引发异常。

        Parameters
        -----------
        func
            要从全局检查中删除的函数。
        call_once: :class:`bool`
            如果函数是在 :meth:`.Bot.add_check` 调用中添加了 ``call_once=True`` 或使用 :meth:`.check_once`。
        """
        l = self._check_once if call_once else self._checks

        try:
            l.remove(func)
        except ValueError:
            pass

    def check_once(self, func: CFT) -> CFT:
        r"""向机器人添加 ``调用一次`` 全局检查的装饰器。

        与常规的全局检查不同，每个 :meth:`.invoke` 调用只调用一次。

        每当调用命令或调用 :meth:`.Command.can_run` 时，都会调用常规的全局检查。
        这种类型的检查绕过它并确保它只被调用一次，即使在默认的帮助命令中也是如此。

        .. note::

            使用此函数时，发送到组子命令的 :class:`.Context` 可能只解析父命令而不解析子命令，因为它在每个 :meth:`.Bot.invoke` 调用中被调用一次。

        .. note::

            此函数可以是常规函数或协程。

        类似于命令 :func:`.check`，它接受一个类型为 :class:`.Context` 的参数，并且只能引发从 :exc:`.CommandError` 继承的异常。

        Example
        ---------

        .. code-block:: python3

            @bot.check_once
            def whitelist(ctx):
                return ctx.message.author.id in my_whitelist

        """
        self.add_check(func, call_once=True)
        return func

    async def can_run(self, ctx: Context, *, call_once: bool = False) -> bool:
        data = self._check_once if call_once else self._checks

        if len(data) == 0:
            return True

        # type-checker doesn't distinguish between functions and methods
        return await qq.utils.async_all(f(ctx) for f in data)  # type: ignore

    async def is_owner(self, user: qq.User) -> bool:
        """|coro|

        检查 :class:`~qq.User` 或 :class:`~qq.Member` 是否是这个机器人的所有者。

        如果 :attr:`owner_id` 未设置，则会通过使用 :meth:`~.Bot.application_info` 自动获取。

        如果 :attr:`owner_ids` 未设置，该函数还会检查应用程序是否为团队所有。

        Parameters
        -----------
        user: :class:`.abc.User`
            要检查的用户。

        Returns
        --------
        :class:`bool`
            用户是否为所有者。
        """

        if self.owner_id:
            return user.id == self.owner_id
        return user.id in self.owner_ids

    def before_invoke(self, coro: CFT) -> CFT:
        """将协程注册为调用前钩的装饰器。

        在调用命令之前直接调用调用前钩。 这使得它成为设置数据库连接或所需的任何类型的设置的有用功能。

        这个调用前钩有一个唯一的参数，一个 :class:`.Context`。

        .. note::

            :meth:`~.Bot.before_invoke` 和 :meth:`~.Bot.after_invoke` 钩子只有在所有检查和参数解析过程无错误通过时才会被调用。
            如果任何检查或参数解析过程失败，则不会调用挂钩。

        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            要注册为调用前钩的协程。

        Raises
        -------
        TypeError
            传递的协程实际上并不是协程。
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('调用前钩必须是一个协程。')

        self._before_invoke = coro
        return coro

    def after_invoke(self, coro: CFT) -> CFT:
        r"""将协程注册为调用后钩的装饰器。

        在调用命令后直接调用调用后钩。 这使其成为清理数据库连接或任何类型的清理所需的有用功能。

        这个调用后钩有一个唯一的参数，一个 :class:`.Context`。

        .. note::

            类似于 :meth:`~.Bot.before_invoke`\，除非检查和参数解析过程成功，否则不会调用它。
            但是，无论内部命令回调是否引发错误（即 :exc:`.CommandInvokeError`\），都会 **始终** 调用此钩子。
            这使其成为清理的理想选择。

        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            要注册为调用后钩的协程。

        Raises
        -------
        TypeError
            传递的协程实际上并不是协程。
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('调用后钩必须是一个协程。')

        self._after_invoke = coro
        return coro

    # listener registration

    def add_listener(self, func: CoroFunc, name: str = MISSING) -> None:
        """:meth:`.listen` 的非装饰器替代品。

        Parameters
        -----------
        func: :ref:`coroutine <coroutine>`
            要调用的函数。
        name: :class:`str`
            要监听的事件的名称。 默认为 ``func.__name__`` 。

        Example
        --------

        .. code-block:: python3

            async def on_ready(): pass
            async def my_message(message): pass

            bot.add_listener(on_ready)
            bot.add_listener(my_message, 'on_message')

        """
        name = func.__name__ if name is MISSING else name

        if not asyncio.iscoroutinefunction(func):
            raise TypeError('监听器必须是协程')

        if name in self.extra_events:
            self.extra_events[name].append(func)
        else:
            self.extra_events[name] = [func]

    def remove_listener(self, func: CoroFunc, name: str = MISSING) -> None:
        """从监听器池中删除一个监听器。

        Parameters
        -----------
        func
            用作要删除的监听器的函数。
        name: :class:`str`
            要删除的事件的名称。 默认为 ``func.__name__`` 。
        """

        name = func.__name__ if name is MISSING else name

        if name in self.extra_events:
            try:
                self.extra_events[name].remove(func)
            except ValueError:
                pass

    def listen(self, name: str = MISSING) -> Callable[[CFT], CFT]:
        """将另一个函数注册为外部事件监听器的装饰器。
        基本上，这允许你收听来自不同地方的多个事件，例如 :func:`.on_ready`

        被监听的函数必须是一个 :ref:`coroutine <coroutine>` 。

        Example
        --------

        .. code-block:: python3

            @bot.listen()
            async def on_message(message):
                print('one')

            # 在其他文件中...

            @bot.listen('on_message')
            async def my_message(message):
                print('two')

        将以未指定的顺序输出一和二。

        Raises
        -------
        TypeError
            被监听的函数不是协程。
        """

        def decorator(func: CFT) -> CFT:
            self.add_listener(func, name)
            return func

        return decorator

    # cogs

    def add_cog(self, cog: Cog, *, override: bool = False) -> None:
        """向机器人添加一个 ``齿轮`` 。

        齿轮是一个拥有自己的事件监听器和命令的类。

        Parameters
        -----------
        cog: :class:`.Cog`
            注册到机器人的齿轮。
        override: :class:`bool`
            如果之前加载的同名 cog 应该被覆盖而不是引发错误。

        Raises
        -------
        TypeError
            齿轮不是从 :class:`.Cog` 继承的。
        CommandError
            加载过程中发生错误。
        .ClientException
            已加载同名的齿轮。
        """

        if not isinstance(cog, Cog):
            raise TypeError('齿轮必须继承自 Cog')

        cog_name = cog.__cog_name__
        existing = self.__cogs.get(cog_name)

        if existing is not None:
            if not override:
                raise qq.ClientException(f'名为 {cog_name!r} 的 Cog 已加载')
            self.remove_cog(cog_name)

        cog = cog._inject(self)
        self.__cogs[cog_name] = cog

    def get_cog(self, name: str) -> Optional[Cog]:
        """获取请求的齿轮实例。

        如果没有找到齿轮，则返回 ``None`` 。

        Parameters
        -----------
        name: :class:`str`
            你请求的齿轮的名称。
            这相当于在类创建中通过关键字参数传递的名称或在未指定参数的时候为类名称。

        Returns
        --------
        Optional[:class:`Cog`]
            请求的齿轮。 如果未找到，则返回 ``None`` 。
        """
        return self.__cogs.get(name)

    def remove_cog(self, name: str) -> Optional[Cog]:
        """从机器人中移除一个齿轮并返回它。

        齿轮注册的所有注册命令和事件监听器也将被删除。

        如果未找到齿轮，则此方法无效。

        Parameters
        -----------
        name: :class:`str`
            要移除的齿轮的名称。

        Returns
        -------
        Optional[:class:`.Cog`]
             被移除的齿轮。 如果未找到，则为 ``None`` 。
        """

        cog = self.__cogs.pop(name, None)
        if cog is None:
            return

        help_command = self._help_command
        if help_command and help_command.cog is cog:
            help_command.cog = None
        cog._eject(self)

        return cog

    @property
    def cogs(self) -> Mapping[str, Cog]:
        """Mapping[:class:`str`, :class:`Cog`]: 齿轮名到齿轮的只读映射。"""
        return types.MappingProxyType(self.__cogs)

    # extensions

    def _remove_module_references(self, name: str) -> None:
        # find all references to the module
        # remove the cogs registered from the module
        for cogname, cog in self.__cogs.copy().items():
            if _is_submodule(name, cog.__module__):
                self.remove_cog(cogname)

        # remove all the commands from the module
        for cmd in self.all_commands.copy().values():
            if cmd.module is not None and _is_submodule(name, cmd.module):
                if isinstance(cmd, GroupMixin):
                    cmd.recursively_remove_all_commands()
                self.remove_command(cmd.name)

        # remove all the listeners from the module
        for event_list in self.extra_events.copy().values():
            remove = []
            for index, event in enumerate(event_list):
                if event.__module__ is not None and _is_submodule(name, event.__module__):
                    remove.append(index)

            for index in reversed(remove):
                del event_list[index]

    def _call_module_finalizers(self, lib: types.ModuleType, key: str) -> None:
        try:
            func = getattr(lib, 'teardown')
        except AttributeError:
            pass
        else:
            try:
                func(self)
            except Exception:
                pass
        finally:
            self.__extensions.pop(key, None)
            sys.modules.pop(key, None)
            name = lib.__name__
            for module in list(sys.modules.keys()):
                if _is_submodule(name, module):
                    del sys.modules[module]

    def _load_from_module_spec(self, spec: importlib.machinery.ModuleSpec, key: str) -> None:
        # precondition: key not in self.__extensions
        lib = importlib.util.module_from_spec(spec)
        sys.modules[key] = lib
        try:
            spec.loader.exec_module(lib)  # type: ignore
        except Exception as e:
            del sys.modules[key]
            raise errors.ExtensionFailed(key, e) from e

        try:
            setup = getattr(lib, 'setup')
        except AttributeError:
            del sys.modules[key]
            raise errors.NoEntryPointError(key)

        try:
            setup(self)
        except Exception as e:
            del sys.modules[key]
            self._remove_module_references(lib.__name__)
            self._call_module_finalizers(lib, key)
            raise errors.ExtensionFailed(key, e) from e
        else:
            self.__extensions[key] = lib

    def _resolve_name(self, name: str, package: Optional[str]) -> str:
        try:
            return importlib.util.resolve_name(name, package)
        except ImportError:
            raise errors.ExtensionNotFound(name)

    def load_extension(self, name: str, *, package: Optional[str] = None) -> None:
        """加载扩展。

        扩展是包含命令、cog 或监听器的 Python 模块。

        扩展必须有一个全局函数，``setup`` 定义为加载扩展时要执行的操作的入口点。 这个入口点必须有一个参数，``bot`` 。

        Parameters
        ------------
        name: :class:`str`
            要加载的扩展名。 如果访问子模块，它必须像常规 Python 导入一样用点分隔。
            例如如果你想导入 ``foo/test.py`` 你可以使用 ``foo.test``  。
        package: Optional[:class:`str`]
            用于解析相对导入的包名。当使用相对路径加载扩展时，这是必需的，例如 ``.foo.test`` 。 默认为 ``None`` 。

        Raises
        --------
        ExtensionNotFound
            无法导入扩展。
            如果无法使用提供的 ``package`` 参数解析扩展的名称，也会引发此问题。
        ExtensionAlreadyLoaded
            扩展已经加载。
        NoEntryPointError
            该扩展没有 ``setup`` 函数。
        ExtensionFailed
            扩展程序或其设置函数有执行错误。
        """

        name = self._resolve_name(name, package)
        if name in self.__extensions:
            raise errors.ExtensionAlreadyLoaded(name)

        spec = importlib.util.find_spec(name)
        if spec is None:
            raise errors.ExtensionNotFound(name)

        self._load_from_module_spec(spec, name)

    def unload_extension(self, name: str, *, package: Optional[str] = None) -> None:
        """卸载扩展。

        卸载扩展后，所有命令、监听器和齿轮都将从机器人中删除，并且模块将取消导入。

        扩展可以提供一个可选的全局函数 ``teardown`` ，以便在必要时进行杂项清理。
        这个函数接受一个参数， ``bot`` ，类似于 :meth:`~.Bot.load_extension` 中的 ``setup`` 。

        Parameters
        ------------
        name: :class:`str`
            要卸载的扩展名。 如果访问子模块，它必须像常规 Python 导入一样用点分隔。
            例如如果你想导入 ``foo/test.py`` 你可以使用 ``foo.test``  。
        package: Optional[:class:`str`]
            用于解析相对导入的包名。当使用相对路径加载扩展时，这是必需的，例如  ``.foo.test``  。 默认为 ``None`` 。

        Raises
        -------
        ExtensionNotFound
            无法导入扩展。
            如果无法使用提供的 ``package`` 参数解析扩展的名称，也会引发此问题。
        ExtensionNotLoaded
            未加载扩展。
        """

        name = self._resolve_name(name, package)
        lib = self.__extensions.get(name)
        if lib is None:
            raise errors.ExtensionNotLoaded(name)

        self._remove_module_references(lib.__name__)
        self._call_module_finalizers(lib, name)

    def reload_extension(self, name: str, *, package: Optional[str] = None) -> None:
        """以原子性的方式重新加载扩展。

        这会用相同的扩展名替换扩展名，只是刷新了而已。
        除了以原子性的方式完成，这相当于一个 :meth:`unload_extension` 后跟一个 meth:`load_extension` 。
        也就是说，如果操作在重新加载过程中失败，那么机器人将回滚到之前的工作状态。

        Parameters
        ------------
        name: :class:`str`
            要重新加载的扩展名。 如果访问子模块，它必须像常规 Python 导入一样用点分隔。
            例如如果你想导入 ``foo/test.py`` 你可以使用 ``foo.test``  。
        package: Optional[:class:`str`]
            用于解析相对导入的包名。当使用相对路径加载扩展时，这是必需的，例如  ``.foo.test``  。 默认为 ``None`` 。

        Raises
        -------
        ExtensionNotLoaded
            未加载扩展。
        ExtensionNotFound
            无法导入扩展。 如果无法使用提供的 ``package`` 参数解析扩展的名称，也会引发此错误。
        NoEntryPointError
            该扩展没有 ``setup`` 函数。
        ExtensionFailed
            扩展 ``setup`` 函数有一个错误。
        """

        name = self._resolve_name(name, package)
        lib = self.__extensions.get(name)
        if lib is None:
            raise errors.ExtensionNotLoaded(name)

        # get the previous module states from sys modules
        modules = {
            name: module
            for name, module in sys.modules.items()
            if _is_submodule(lib.__name__, name)
        }

        try:
            # Unload and then load the module...
            self._remove_module_references(lib.__name__)
            self._call_module_finalizers(lib, name)
            self.load_extension(name)
        except Exception:
            # if the load failed, the remnants should have been
            # cleaned from the load_extension function call
            # so let's load it from our old compiled library.
            lib.setup(self)  # type: ignore
            self.__extensions[name] = lib

            # revert sys.modules back to normal and raise back to caller
            sys.modules.update(modules)
            raise

    @property
    def extensions(self) -> Mapping[str, types.ModuleType]:
        """Mapping[:class:`str`, :class:`py:types.ModuleType`]: 扩展名到扩展的只读映射。"""
        return types.MappingProxyType(self.__extensions)

    # help command stuff

    @property
    def help_command(self) -> Optional[HelpCommand]:
        return self._help_command

    @help_command.setter
    def help_command(self, value: Optional[HelpCommand]) -> None:
        if value is not None:
            if not isinstance(value, HelpCommand):
                raise TypeError('help_command must be a subclass of HelpCommand')
            if self._help_command is not None:
                self._help_command._remove_from_bot(self)
            self._help_command = value
            value._add_to_bot(self)
        elif self._help_command is not None:
            self._help_command._remove_from_bot(self)
            self._help_command = None
        else:
            self._help_command = None

    # command processing

    async def get_prefix(self, message: Message) -> Union[List[str], str]:
        """|coro|

        检索机器人正在监听的前缀，并将消息作为 context。

        Parameters
        -----------
        message: :class:`qq.Message`
            要获取前缀的 context 。

        Returns
        --------
        Union[List[:class:`str`], :class:`str`]
            机器人正在监听的前缀列表或单个前缀。
        """
        prefix = ret = self.command_prefix
        if callable(prefix):
            ret = await qq.utils.maybe_coroutine(prefix, self, message)

        if not isinstance(ret, str):
            try:
                ret = list(ret)
            except TypeError:
                # It's possible that a generator raised this exception.  Don't
                # replace it with our own error if that's the case.
                if isinstance(ret, collections.abc.Iterable):
                    raise

                raise TypeError("command_prefix must be plain string, iterable of strings, or callable "
                                f"returning either of these, not {ret.__class__.__name__}")

            if not ret:
                raise ValueError("Iterable command_prefix must contain at least one prefix")

        return ret

    async def get_context(self, message: Message, *, cls: Type[CXT] = Context) -> CXT:
        r"""|coro|

        从消息中返回调用context。

        这是 :meth:`.process_commands` 的一个更底层的对应部分，允许用户对处理进行更细度的控制。

        返回的 context 不能保证是有效的 调用context，必须检查 :attr:`.Context.valid` 以确保它是有效的。
        如果 context 无效，则它不会在 :meth:`~.Bot.invoke` 下调用。

        Parameters
        -----------
        message: :class:`qq.Message`
            从中获取调用 context 的消息。
        cls
            将用于创建 context 的工厂类。默认情况下，这是 :class:`.Context`。
            如果提供自定义类，它必须与 :class:`.Context`\ 的接口足够相似。

        Returns
        --------
        :class:`.Context`
            调用 context 。 这个类型可以通过 ``cls`` 参数改变。
        """

        view = StringView(message.content)
        ctx = cls(prefix=None, view=view, bot=self, message=message)

        if message.author.id == self.user.id:  # type: ignore
            return ctx

        prefix = await self.get_prefix(message)
        invoked_prefix = prefix

        if isinstance(prefix, str):
            if not view.skip_string(prefix):
                return ctx
        else:
            try:
                # if the context class' __init__ consumes something from the view this
                # will be wrong.  That seems unreasonable though.
                if message.content.startswith(tuple(prefix)):
                    invoked_prefix = qq.utils.find(view.skip_string, prefix)
                else:
                    return ctx

            except TypeError:
                if not isinstance(prefix, list):
                    raise TypeError("get_prefix 必须返回字符串或字符串列表，"
                                    f"而不是 {prefix.__class__.__name__}")

                # It's possible a bad command_prefix got us here.
                for value in prefix:
                    if not isinstance(value, str):
                        raise TypeError("从 get_prefix 返回的可迭代 command_prefix 或列表必须只包含字符串，"
                                        f"而不是 {value.__class__.__name__}")

                # Getting here shouldn't happen
                raise

        if self.strip_after_prefix:
            view.skip_ws()

        invoker = view.get_word()
        ctx.invoked_with = invoker
        # type-checker fails to narrow invoked_prefix type.
        ctx.prefix = invoked_prefix  # type: ignore
        ctx.command = self.all_commands.get(invoker)
        return ctx

    async def invoke(self, ctx: Context) -> None:
        """|coro|

        调用在调用 context 下给出的命令并处理所有内部事件调度机制。

        Parameters
        -----------
        ctx: :class:`.Context`
            要调用的调用 context 。
        """
        if ctx.command is not None:
            self.dispatch('command', ctx)
            try:
                if await self.can_run(ctx, call_once=True):
                    await ctx.command.invoke(ctx)
                else:
                    raise errors.CheckFailure('The global check once functions failed.')
            except errors.CommandError as exc:
                await ctx.command.dispatch_error(ctx, exc)
            else:
                self.dispatch('command_completion', ctx)
        elif ctx.invoked_with:
            exc = errors.CommandNotFound(f'Command "{ctx.invoked_with}" is not found')
            self.dispatch('command_error', ctx, exc)

    async def process_commands(self, message: Message) -> None:
        """|coro|

        函数处理已注册到机器人和其他组的命令。 如果没有这个协程，将不会触发任何命令。

        默认情况下，这个协程在 :func:`.on_message` 事件中被调用。
        如果你选择覆盖 :func:`.on_message` 事件，那么你也应该调用这个协程。

        这是使用其他低级工具构建的，
        相当于调用 :meth:`~.Bot.get_context` 然后调用 :meth:`~.Bot.invoke`。

        这还会检查消息的作者是否是机器人，如果是，则不会调用 :meth:`~.Bot.get_context` 或 :meth:`~.Bot.invoke`。

        Parameters
        -----------
        message: :class:`qq.Message`
            要为其处理命令的消息。
        """
        if message.author.bot:
            return

        ctx = await self.get_context(message)
        await self.invoke(ctx)

    async def on_message(self, message):
        await self.process_commands(message)


class Bot(BotBase, qq.Client):
    """代表一个qq机器人。

    这个类是 :class:`qq.Client` 的子类，因此你可以用这个机器人做任何你可以 :class:`qq.Client` 用做的事情。

    该类还继承了 :class:`.GroupMixin` 以提供管理命令的功能。

    Attributes
    -----------
    command_prefix
        命令前缀是消息内容最初必须包含的内容才能调用命令。
        这个前缀可以是一个字符串来指示前缀应该是什么，
        或者是一个可调用的，它接受机器人作为它的第一个参数和 :class:`qq.Message` 作为它的第二个参数并返回前缀。
        这是为了方便“动态`` 命令前缀。 这个可调用对象可以是常规函数或协程。

        空字符串作为前缀将始终匹配，启用无前缀命令调用。
        虽然这在私聊中可能很有用，但在服务器中应该避免，因为它可能会导致性能问题和意外的命令调用。

        命令前缀也可以是一个可迭代的字符串，表示应该对前缀进行多次检查，第一个匹配的将是调用前缀。
        你可以通过 :attr:`.Context.prefix` 得到这个前缀。 为避免混淆，不允许使用空的可迭代对象。

        .. note::

            传递多个前缀时，请注意不要传递与序列中后面出现的较长前缀相匹配的前缀。
            例如，如果命令前缀是 ``('!', '!?')`` ``'!?'`` 前缀将永远不会与任何消息匹配，因为前一个匹配 ``!`` 是后一个的子集 。
            这在传递空字符串时尤其重要，它应该始终放在最后，因为在匹配之后没有前缀能够被匹配。

    case_insensitive: :class:`bool`
        命令是否应该不区分大小写。 默认为 ``False`` 。 此属性不会延续到组。 如果你还要求组命令不区分大小写，则必须将其设置为每个组。
    description: :class:`str`
        默认帮助消息中带有前缀的内容。
    help_command: Optional[:class:`.HelpCommand`]
        要使用的帮助命令实现。
        这可以在运行时动态设置。 删除帮助命令传递 ``None`` 。
        有关实现帮助命令的更多信息，请参阅 :ref:`ext_commands_help_command`。
    owner_id: Optional[:class:`int`]
        拥有机器人的用户 ID。 必须设置 ``owner_id`` 或 ``owner_ids`` 其中一个
    owner_ids: Optional[Collection[:class:`int`]]
        拥有机器人的多个用户 ID。 这类似于 :attr:`owner_id`。
        出于性能原因，建议使用 :class:`set`。 必须设置 ``owner_id`` 或 ``owner_ids`` 其中一个
    strip_after_prefix: :class:`bool`
        遇到命令前缀后是否去除空白字符。 这允许 ``!   hello`` 和 ``!hello`` 都生效，默认为 ''False''。
    """
    pass


class AutoShardedBot(BotBase, qq.AutoShardedClient):
    """除了它是从 :class:`qq.AutoShardedClient` 继承的，其他与 :class:`.Bot` 类似。
    """
    pass



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import datetime
from typing import Any, Dict, Final, List, Mapping, Protocol, TYPE_CHECKING, Type, TypeVar, Union, Optional

__all__ = (
    'Embed',
    'Ark'
)

from . import utils
from .colour import Colour


class _EmptyEmbed:
    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return 'Embed.Empty'

    def __len__(self) -> int:
        return 0


EmptyEmbed: Final = _EmptyEmbed()


class EmbedProxy:
    def __init__(self, layer: Dict[str, Any]):
        self.__dict__.update(layer)

    def __len__(self) -> int:
        return len(self.__dict__)

    def __repr__(self) -> str:
        inner = ', '.join((f'{k}={v!r}' for k, v in self.__dict__.items() if not k.startswith('_')))
        return f'EmbedProxy({inner})'

    def __getattr__(self, attr: str) -> _EmptyEmbed:
        return EmptyEmbed


E = TypeVar('E', bound='Embed')

if TYPE_CHECKING:
    from .types.embed import Embed as EmbedData

    T = TypeVar('T')
    MaybeEmpty = Union[T, _EmptyEmbed]


    class _EmbedFooterProxy(Protocol):
        text: MaybeEmpty[str]
        icon_url: MaybeEmpty[str]


    class _EmbedFieldProxy(Protocol):
        name: MaybeEmpty[str]
        value: MaybeEmpty[str]
        inline: bool


    class _EmbedMediaProxy(Protocol):
        url: MaybeEmpty[str]
        proxy_url: MaybeEmpty[str]
        height: MaybeEmpty[int]
        width: MaybeEmpty[int]


    class _EmbedVideoProxy(Protocol):
        url: MaybeEmpty[str]
        height: MaybeEmpty[int]
        width: MaybeEmpty[int]


    class _EmbedProviderProxy(Protocol):
        name: MaybeEmpty[str]
        url: MaybeEmpty[str]


    class _EmbedAuthorProxy(Protocol):
        name: MaybeEmpty[str]
        url: MaybeEmpty[str]
        icon_url: MaybeEmpty[str]
        proxy_icon_url: MaybeEmpty[str]


class Ark:
    """代表一个 QQ Ark。

    .. container:: operations

        .. describe:: len(x)

            返回嵌入的总大小。用于检查它是否在 6000 个字符限制内。

        .. describe:: bool(b)

            返回嵌入是否有任何数据集。

    Attributes
    -----------
    template_id: :class:`str`
        要使用的模版 ID。
    colour: Union[:class:`Colour`, :class:`int`]
        嵌入的颜色代码。别名为 ``color``。这可以在初始化期间设置。目前官方还没有实现
    """
    __slots__ = (
        '_fields',
        '_colour',
        'template_id',
        'colour',
        '_extra'
    )

    def __init__(
            self,
            *,
            colour: Optional[Union[int, Colour, _EmptyEmbed]] = EmptyEmbed,
            color: Optional[Union[int, Colour, _EmptyEmbed]] = EmptyEmbed,
            template_id: int,
    ):
        self._extra = {}
        self.template_id = template_id
        self.colour = colour if colour is not EmptyEmbed else color if color is not EmptyEmbed else None

    @property
    def fields(self) -> List[_EmbedFieldProxy]:
        """List[Union[``EmbedProxy``, :attr:`Empty`]]:
        返回 ``EmbedProxy`` 的一个列表，表示字段内容。有关您可以访问的可能值，请参阅 :meth:`add_field`。
        如果该属性没有值，则返回 :attr:`Empty`。
        """
        return [EmbedProxy(d) for d in getattr(self, '_fields', [])]  # type: ignore

    def set_attribute(self: E, key: Any, value: Any) -> E:
        """向 Ark 对象添加字段。此函数返回类实例以允许流式链接。

        Parameters
        -----------
        key: :class:`str`
            字段的名称。
        value: :class:`str`
            字段的值。
        """

        self._extra[str(key)] = str(value)
        return self

    def add_field(self: E, *, desc: Any, url: Any = None) -> E:
        """向 Ark 对象添加字段。此函数返回类实例以允许流式链接。

        Parameters
        -----------
        desc: :class:`str`
            字段的值。
        url: :class:`str`
            字段的 Url。
        """

        field = {'obj_kv': []}
        if desc:
            field['obj_kv'].append({
                "key": "desc",
                "value": str(desc)
            })
        if url:
            field['obj_kv'].append({
                "key": "url",
                "value": str(desc)
            })

        try:
            self._fields.append(field)  # type: ignore
        except AttributeError:
            self._fields = [field]

        return self

    def insert_field_at(self: E, index: int, *, desc: Any, url: Any = None) -> E:
        """在 Ark 的指定索引之前插入一个字段。此函数返回类实例以允许流式链接。

        Parameters
        -----------
        index: :class:`int`
            插入字段的位置的索引。
        desc: :class:`str`
            字段的值。
        url: :class:`str`
            字段的 Url。
        """

        field = {'obj_kv': []}
        if desc:
            field['obj_kv'].append({
                "key": "desc",
                "value": str(desc)
            })
        if url:
            field['obj_kv'].append({
                "key": "link",
                "value": str(desc)
            })

        try:
            self._fields.insert(index, field)  # type: ignore
        except AttributeError:
            self._fields = [field]

        return self

    def clear_fields(self) -> None:
        """从此 Ark 中删除所有字段。"""
        try:
            self._fields.clear()
        except AttributeError:
            self._fields = []

    def to_dict(self) -> EmbedData:
        """将此 Ark 对象转换为字典。"""

        result = {"template_id": self.template_id,
                  "kv": [{"key": i, "value": j} for i, j in self._extra.items()]}

        if self.fields:
            result["kv"].append({"key": "#LIST#", "obj": self._fields})  # type: ignore

        return result  # type: ignore


class Embed:
    """代表一个 QQ 嵌入。

    .. container:: operations

        .. describe:: len(x)

            返回嵌入的总大小。用于检查它是否在 6000 个字符限制内。

        .. describe:: bool(b)

            返回嵌入是否有任何数据集。

    某些属性返回一个 ``EmbedProxy`` ，
    除了使用点访问，该类型的行为类似于常规 :class:`dict`，例如 ``embed.author.icon_url``。
    如果属性无效或为空，则返回一个特殊的值，:attr:`Embed.Empty`。

    为了便于使用，所有需要 :class:`str` 的参数都为您隐式转换为 :class:`str`。

    Attributes
    -----------
    title: :class:`str`
        嵌入的标题。这可以在初始化期间设置。
    description: :class:`str`
        嵌入的描述。这可以在初始化期间设置。
    timestamp: :class:`datetime.datetime`
        嵌入内容的时间戳。这是一个 aware 的 datetime。如果传递了一个简单的 datetime，它会被转换为具有本地时区的已知 datetime。
    colour: Union[:class:`Colour`, :class:`int`]
        嵌入的颜色代码。别名为 ``color``。这可以在初始化期间设置。目前官方还没有实现
    prompt: :class: `str`
        嵌入的弹出窗口。这可以在初始化期间设置。
    """
    __slots__ = (
        'title',
        '_fields',
        'description',
        'prompt',
        '_fields',
        '_timestamp',
        '_author',
        '_footer',
        '_colour'
    )

    Empty: Final = EmptyEmbed

    def __init__(
            self,
            *,
            title: MaybeEmpty[Any] = EmptyEmbed,
            colour: Optional[Union[int, Colour, _EmptyEmbed]] = EmptyEmbed,
            color: Optional[Union[int, Colour, _EmptyEmbed]] = EmptyEmbed,
            description: MaybeEmpty[Any] = EmptyEmbed,
            prompt: MaybeEmpty[Any] = EmptyEmbed,
            timestamp: datetime.datetime = None,
    ):
        self.colour = colour if colour is not EmptyEmbed else color if color is not EmptyEmbed else None
        self.title = title
        self.description = description
        self.prompt = prompt

        if self.prompt is not EmptyEmbed:
            self.prompt = str(self.prompt)

        if self.title is not EmptyEmbed:
            self.title = str(self.title)

        if self.description is not EmptyEmbed:
            self.description = str(self.description)

        if timestamp:
            self.timestamp = timestamp

    @classmethod
    def from_dict(cls: Type[E], data: Mapping[str, Any]) -> E:
        """将 :class:`dict` 转换为 :class:`Embed`，前提是它采用 QQ 期望的格式。
        您可以在 ``官方 QQ 文档``__ 中找到有关此格式的信息。

        .. _QQDocs: https://bot.q.qq.com/wiki/develop/api/openapi/message/model.html#messageembed
        __ QQDocs_

        Parameters
        -----------
        data: :class:`dict`
            要转换为嵌入的字典。
        """
        # we are bypassing __init__ here since it doesn't apply here
        self: E = cls.__new__(cls)

        # fill in the basic fields

        self.title = data.get('title', EmptyEmbed)
        self.description = data.get('description', EmptyEmbed)
        self.prompt = data.get('prompt', EmptyEmbed)

        if self.title is not EmptyEmbed:
            self.title = str(self.title)

        if self.description is not EmptyEmbed:
            self.description = str(self.description)

        # try to fill in the more rich fields

        try:
            self._colour = Colour(value=int(data['color']))
        except KeyError:
            pass

        try:
            self._timestamp = utils.parse_time(data['timestamp'])
        except KeyError:
            pass

        for attr in ('thumbnail', 'video', 'provider', 'author', 'fields', 'image', 'footer'):
            try:
                value = data[attr]
            except KeyError:
                continue
            else:
                setattr(self, '_' + attr, value)

        return self

    def copy(self: E) -> E:
        """返回嵌入的浅表副本。"""
        return self.__class__.from_dict(self.to_dict())

    def __len__(self) -> int:
        total = len(self.title) + len(self.description)
        for field in getattr(self, '_fields', []):
            total += len(field['name']) + len(field['value'])

        try:
            footer_text = self._footer['text']
        except (AttributeError, KeyError):
            pass
        else:
            total += len(footer_text)

        try:
            author = self._author
        except AttributeError:
            pass
        else:
            total += len(author['name'])

        return total

    def __bool__(self) -> bool:
        return any(
            (
                self.title,
                self.description,
                self.colour,
                self.fields,
                self.timestamp,
                self.author,
                self.thumbnail,
                self.footer,
                self.image,
                self.provider,
                self.video,
            )
        )

    @property
    def colour(self) -> MaybeEmpty[Colour]:
        return getattr(self, '_colour', EmptyEmbed)

    @colour.setter
    def colour(self, value: Union[int, Colour, _EmptyEmbed]):  # type: ignore
        if isinstance(value, (Colour, _EmptyEmbed)):
            self._colour = value
        elif isinstance(value, int):
            self._colour = Colour(value=value)
        else:
            raise TypeError(
                f'Expected qq.Colour, int, or Embed.Empty but received {value.__class__.__name__} instead.')

    color = colour

    @property
    def timestamp(self) -> MaybeEmpty[datetime.datetime]:
        return getattr(self, '_timestamp', EmptyEmbed)

    @timestamp.setter
    def timestamp(self, value: MaybeEmpty[datetime.datetime]):
        if isinstance(value, datetime.datetime):
            if value.tzinfo is None:
                value = value.astimezone()
            self._timestamp = value
        elif isinstance(value, _EmptyEmbed):
            self._timestamp = value
        else:
            raise TypeError(
                f"Expected datetime.datetime or Embed.Empty received {value.__class__.__name__} instead")

    @property
    def footer(self) -> _EmbedFooterProxy:
        """返回表示页脚内容的 ``EmbedProxy``。
        请参阅 :meth:`set_footer` 以获取您可以访问的可能值。
        如果该属性没有值，则返回 :attr:`Empty`。


        """
        return EmbedProxy(getattr(self, '_footer', {}))  # type: ignore

    def set_footer(self: E, *, text: MaybeEmpty[Any] = EmptyEmbed, icon_url: MaybeEmpty[Any] = EmptyEmbed) -> E:
        """设置嵌入内容的页脚。此函数返回类实例以允许流式链接。



        Parameters
        -----------
        text: :class:`str`
            页脚文本。
        icon_url: :class:`str`
            页脚图标的 URL。仅支持 HTTP(S)。
        """

        self._footer = {}
        if text is not EmptyEmbed:
            self._footer['text'] = str(text)

        if icon_url is not EmptyEmbed:
            self._footer['icon_url'] = str(icon_url)

        return self

    def remove_footer(self: E) -> E:
        """清除嵌入的页脚信息。此函数返回类实例以允许流式链接。


        """
        try:
            del self._footer
        except AttributeError:
            pass

        return self

    @property
    def image(self) -> _EmbedMediaProxy:
        """返回表示图像内容的 ``EmbedProxy`` 。

        您可以访问的可能属性是：

        - ``url``
        - ``proxy_url``
        - ``width``
        - ``height``

        如果该属性没有值，则返回 :attr:`Empty`。


        """
        return EmbedProxy(getattr(self, '_image', {}))  # type: ignore

    def set_image(self: E, *, url: MaybeEmpty[Any]) -> E:
        """设置嵌入内容的图像。
        此函数返回类实例以允许流式链接。
        传递 :attr:`Empty` 会删除图像。



        Parameters
        -----------
        url: :class:`str`
            图像的源 URL。仅支持 HTTP(S)。
        """

        if url is EmptyEmbed:
            try:
                del self._image
            except AttributeError:
                pass
        else:
            self._image = {
                'url': str(url),
            }

        return self

    @property
    def thumbnail(self) -> _EmbedMediaProxy:
        """返回表示缩略图内容的 ``EmbedProxy`` 。

        您可以访问的可能属性是：

        - ``url``
        - ``proxy_url``
        - ``width``
        - ``height``

        如果该属性没有值，则返回 :attr:`Empty`。


        """
        return EmbedProxy(getattr(self, '_thumbnail', {}))  # type: ignore

    def set_thumbnail(self: E, *, url: MaybeEmpty[Any]) -> E:
        """设置嵌入内容的缩略图。此函数返回类实例以允许流式链接。传递 :attr:`Empty` 会删除缩略图。

        Parameters
        -----------
        url: :class:`str`
            缩略图的源 URL。仅支持 HTTP(S)。


        """

        if url is EmptyEmbed:
            try:
                del self._thumbnail
            except AttributeError:
                pass
        else:
            self._thumbnail = {
                'url': str(url),
            }

        return self

    @property
    def video(self) -> _EmbedVideoProxy:
        """返回表示视频内容的 ``EmbedProxy`` 。

        可能的属性包括：

        - ``url`` 视频网址。
        - ``height`` 视频高度。
        - ``width`` 视频宽度。

        如果该属性没有值，则返回 :attr:`Empty`。


        """
        return EmbedProxy(getattr(self, '_video', {}))  # type: ignore

    @property
    def provider(self) -> _EmbedProviderProxy:
        """返回表示提供者内容的 ``EmbedProxy`` 。
        可能被访问的唯一属性是 ``name`` 和 ``url`` 。如果该属性没有值，则返回 :attr:`Empty`。


        """
        return EmbedProxy(getattr(self, '_provider', {}))  # type: ignore

    @property
    def author(self) -> _EmbedAuthorProxy:
        """返回表示作者内容的 ``EmbedProxy`` 。有关您可以访问的可能值，请参阅 :meth:`set_author`。
        如果该属性没有值，则返回 :attr:`Empty`。


        """
        return EmbedProxy(getattr(self, '_author', {}))  # type: ignore

    def set_author(self: E, *, name: Any, url: MaybeEmpty[Any] = EmptyEmbed,
                   icon_url: MaybeEmpty[Any] = EmptyEmbed) -> E:
        """设置嵌入内容的作者。此函数返回类实例以允许流式链接。

        Parameters
        -----------
        name: :class:`str`
            作者的名字。
        url: :class:`str`
            作者的 URL。
        icon_url: :class:`str`
            作者图标的 URL。仅支持 HTTP(S)。


        """

        self._author = {
            'name': str(name),
        }

        if url is not EmptyEmbed:
            self._author['url'] = str(url)

        if icon_url is not EmptyEmbed:
            self._author['icon_url'] = str(icon_url)

        return self

    def remove_author(self: E) -> E:
        """清除嵌入的作者信息。此函数返回类实例以允许流式链接。
        """
        try:
            del self._author
        except AttributeError:
            pass

        return self

    @property
    def fields(self) -> List[_EmbedFieldProxy]:
        """List[Union[``EmbedProxy``, :attr:`Empty`]]:
        返回 ``EmbedProxy`` 的一个列表，表示字段内容。有关您可以访问的可能值，请参阅 :meth:`add_field`。
        如果该属性没有值，则返回 :attr:`Empty`。
        """
        return [EmbedProxy(d) for d in getattr(self, '_fields', [])]  # type: ignore

    def add_field(self: E, *, name: str, value: Optional[str] = '', inline: bool = True) -> E:
        """向嵌入对象添加字段。此函数返回类实例以允许流式链接。

        Parameters
        -----------
        name: :class:`str`
            字段的名称。
        value: :class:`str`
            字段的值。目前并无作用
        inline: :class:`bool`
            该字段是否应内联显示。目前并无作用
        """

        field = {
            'inline': inline,
            'name': str(name),
            'value': str(value),
        }

        try:
            self._fields.append(field)
        except AttributeError:
            self._fields = [field]

        return self

    def insert_field_at(self: E, index: int, *, name: str, value: Optional[str] = '', inline: bool = True) -> E:
        """在嵌入的指定索引之前插入一个字段。此函数返回类实例以允许流式链接。

        Parameters
        -----------
        index: :class:`int`
            插入字段的位置的索引。
        name: :class:`str`
            字段的名称。
        value: :class:`str`
            字段的值。目前并无作用
        inline: :class:`bool`
            该字段是否应内联显示。目前并无作用
        """

        field = {
            'inline': inline,
            'name': str(name),
            'value': str(value),
        }

        try:
            self._fields.insert(index, field)
        except AttributeError:
            self._fields = [field]

        return self

    def clear_fields(self) -> None:
        """从此嵌入中删除所有字段。"""
        try:
            self._fields.clear()
        except AttributeError:
            self._fields = []

    def remove_field(self, index: int) -> None:
        """删除指定索引处的字段。
        如果索引无效或越界，则错误会被默默吞下。

        .. note::

            当按索引删除一个字段时，其他字段的索引会移动以填补空白，就像常规列表一样。

        Parameters
        -----------
        index: :class:`int`
            要删除的字段的索引。
        """
        try:
            del self._fields[index]
        except (AttributeError, IndexError):
            pass

    def set_field_at(self: E, index: int, *, name: Any, value: Any, inline: bool = True) -> E:
        """修改嵌入对象的字段。索引必须指向一个有效的预先存在的字段。此函数返回类实例以允许流式链接。

        Parameters
        -----------
        index: :class:`int`
            要修改的字段的索引。
        name: :class:`str`
            字段的名称。
        value: :class:`str`
            字段的值。
        inline: :class:`bool`
            该字段是否应内联显示。

        Raises
        -------
        IndexError
            提供了无效的索引。
        """

        try:
            field = self._fields[index]
        except (TypeError, IndexError, AttributeError):
            raise IndexError('field index out of range')

        field['name'] = str(name)
        field['value'] = str(value)
        field['inline'] = inline
        return self

    def to_dict(self) -> EmbedData:
        """将此嵌入对象转换为字典。"""

        # add in the raw data into the dict
        # fmt: off
        result = {
            key[1:]: getattr(self, key)
            for key in self.__slots__
            if key[0] == '_' and hasattr(self, key)
        }
        # fmt: on

        # deal with basic convenience wrappers

        try:
            colour = result.pop('colour')
        except KeyError:
            pass
        else:
            if colour:
                result['color'] = colour.value

        try:
            timestamp = result.pop('timestamp')
        except KeyError:
            pass
        else:
            if timestamp:
                if timestamp.tzinfo:
                    result['timestamp'] = timestamp.astimezone(tz=datetime.timezone.utc).isoformat()
                else:
                    result['timestamp'] = timestamp.replace(tzinfo=datetime.timezone.utc).isoformat()

        if self.description:
            result['description'] = self.description

        if self.prompt:
            result['prompt'] = self.prompt

        if self.title:
            result['title'] = self.title

        return result  # type: ignore



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

import array
import asyncio
import datetime
import json
import re
import sys
import unicodedata
from bisect import bisect_left
from operator import attrgetter
from typing import Any, Callable, TypeVar, overload, Optional, Iterable, List, TYPE_CHECKING, Generic, Type, Dict, \
    ForwardRef, Literal, Tuple, Union, Iterator, AsyncIterator
from inspect import isawaitable as _isawaitable, signature as _signature

T = TypeVar('T')
T_co = TypeVar('T_co', covariant=True)
_Iter = Union[Iterator[T], AsyncIterator[T]]
_MARKDOWN_ESCAPE_SUBREGEX = '|'.join(r'\{0}(?=([\s\S]*((?<!\{0})\{0})))'.format(c) for c in ('*', '`', '_', '~', '|'))
_MARKDOWN_ESCAPE_COMMON = r'^>(?:>>)?\s|\[.+\]\(.+\)'
_MARKDOWN_ESCAPE_REGEX = re.compile(fr'(?P<markdown>{_MARKDOWN_ESCAPE_SUBREGEX}|{_MARKDOWN_ESCAPE_COMMON})',
                                    re.MULTILINE)
_URL_REGEX = r'(?P<url><[^: >]+:\/[^ >]+>|(?:https?|steam):\/\/[^\s<]+[^<.,:;\"\'\]\s])'
_MARKDOWN_STOCK_REGEX = fr'(?P<markdown>[_\\~|\*`]|{_MARKDOWN_ESCAPE_COMMON})'

__all__ = (
    'find',
    'get',
    'sleep_until',
    'utcnow',
    'remove_markdown',
    'escape_markdown',
    'escape_mentions',
    'as_chunks',
    'format_dt',
)

_IS_ASCII = re.compile(r'^[\x00-\x7f]+$')


def _string_width(string: str, *, _IS_ASCII=_IS_ASCII) -> int:
    """Returns string's width."""
    match = _IS_ASCII.match(string)
    if match:
        return match.endpos

    UNICODE_WIDE_CHAR_TYPE = 'WFA'
    func = unicodedata.east_asian_width
    return sum(2 if func(char) in UNICODE_WIDE_CHAR_TYPE else 1 for char in string)


def utcnow() -> datetime.datetime:
    """一个辅助函数，用于返回表示当前时间的 UTC datetime。
    这应该比 :meth:`datetime.datetime.utcnow` 更可取，因为与标准库中的原始日期时间相比，它是一个 aware 的 datetime。

    Returns
    --------
    :class:`datetime.datetime`
        UTC 中的当前 datetime datetime。
    """
    return datetime.datetime.now(datetime.timezone.utc)


def _chunk(iterator: Iterator[T], max_size: int) -> Iterator[List[T]]:
    ret = []
    n = 0
    for item in iterator:
        ret.append(item)
        n += 1
        if n == max_size:
            yield ret
            ret = []
            n = 0
    if ret:
        yield ret


async def _achunk(iterator: AsyncIterator[T], max_size: int) -> AsyncIterator[List[T]]:
    ret = []
    n = 0
    async for item in iterator:
        ret.append(item)
        n += 1
        if n == max_size:
            yield ret
            ret = []
            n = 0
    if ret:
        yield ret


@overload
def as_chunks(iterator: Iterator[T], max_size: int) -> Iterator[List[T]]:
    ...


@overload
def as_chunks(iterator: AsyncIterator[T], max_size: int) -> AsyncIterator[List[T]]:
    ...


def as_chunks(iterator: _Iter[T], max_size: int) -> _Iter[List[T]]:
    """将迭代器收集到给定大小的块中的辅助函数。

    Parameters
    ----------
    iterator: Union[:class:`collections.abc.Iterator`, :class:`collections.abc.AsyncIterator`]
        块的迭代器，可以是同步的或异步的。
    max_size: :class:`int`
        最大块大小。


    .. warning::

        收集的最后一个块可能没有 ``max_size`` 那么大。

    Returns
    --------
    Union[:class:`Iterator`, :class:`AsyncIterator`]
        一个新的迭代器，它产生给定大小的块。
    """

    if max_size <= 0:
        raise ValueError('Chunk sizes must be greater than 0.')

    if isinstance(iterator, AsyncIterator):
        return _achunk(iterator, max_size)
    return _chunk(iterator, max_size)


PY_310 = sys.version_info >= (3, 10)
TimestampStyle = Literal['f', 'F', 'd', 'D', 't', 'T', 'R']


def format_dt(dt: datetime.datetime, /, style: Optional[TimestampStyle] = None) -> str:
    """用于格式化 datetime.以在 QQ 中展示的辅助函数。

    +------+----------------------------+------------+
    | 样式 | 示例输出                   | 描述       |
    +------+----------------------------+------------+
    | t    | 22:57                      | 短时间     |
    +------+----------------------------+------------+
    | T    | 22:57:58                   | 长时间     |
    +------+----------------------------+------------+
    | d    | 17/05/2016                 | 短日期     |
    +------+----------------------------+------------+
    | D    | 17 May 2016                | 长日期     |
    +------+----------------------------+------------+
    | f    | 17 May 2016 22:57          | 短日期时间 |
    +------+----------------------------+------------+
    | F    | Tuesday, 17 May 2016 22:57 | 长日期时间 |
    +------+----------------------------+------------+
    | R    | 5 years ago                | 相对时间   |
    +------+----------------------------+------------+

    请注意，确切的输出取决于客户端中用户的区域设置。显示的示例输出使用 ``en-GB`` 语言环境。

    Parameters
    -----------
    dt: :class:`datetime.datetime`
        要格式化的 datetime 。
    style: :class:`str`
        格式化日期时间的样式。

    Returns
    --------
    :class:`str`
        格式化的字符串。
    """
    if style is None:
        return f'<t:{int(dt.timestamp())}>'
    return f'<t:{int(dt.timestamp())}:{style}>'


def compute_timedelta(dt: datetime.datetime):
    if dt.tzinfo is None:
        dt = dt.astimezone()
    now = datetime.datetime.now(datetime.timezone.utc)
    return max((dt - now).total_seconds(), 0)


async def sleep_until(when: datetime.datetime, result: Optional[T] = None) -> Optional[T]:
    """|coro|
    睡眠到指定时间。
    如果提供的时间在过去，则此函数将立即返回。

    Parameters
    -----------
    when: :class:`datetime.datetime`
        休眠到的时间戳。如果日期时间是 native 的，那么它被假定为本地时间。
    result: Any
        如果提供，则在协程完成时返回给调用者。
    """
    delta = compute_timedelta(when)
    return await asyncio.sleep(delta, result)


def remove_markdown(text: str, *, ignore_links: bool = True) -> str:
    """删除 Markdown 字符的辅助函数。

    .. note::
    
            此功能不会解析 Markdown ，可能会从原文中删除含义。 例如，
            如果输入包含 ``10 * 5`` ，那么它将被转换为 ``10 5`` 。

    Parameters
    -----------
    text: :class:`str`
        要从中删除Markdown的文本。
    ignore_links: :class:`bool`
        删除 Markdown 时是否留下链接。 例如，如果文本中的 URL 包含诸如 ``_`` 之类的字符，则它将单独保留。默认为 ``True`` 。

    Returns
    --------
    :class:`str`
        删除了 Markdown 特殊字符的文本。
    """
    def replacement(match):
        groupdict = match.groupdict()
        return groupdict.get('url', '')

    regex = _MARKDOWN_STOCK_REGEX
    if ignore_links:
        regex = f'(?:{_URL_REGEX}|{regex})'
    return re.sub(regex, replacement, text, 0, re.MULTILINE)


def escape_markdown(text: str, *, as_needed: bool = False, ignore_links: bool = True) -> str:
    r"""转义 Markdown 的辅助函数。

    Parameters
    -----------
    text: :class:`str`
        转义 markdown 的文本。
    as_needed: :class:`bool`
        是否根据需要转义Markdown字符。
        这意味着如果没有必要，它不会转义无关的字符，例如 ``hello`` 转义为 ``\\hello`` 而不是 ``\\hello\\`` 。
        但是请注意，这可能会让你面临一些奇怪的语法滥用。默认为 ``False`` 。

    ignore_links: :class:`bool`
        转义 markdown 时是否留下链接。例如，如果文本中的 URL 包含诸如 ``_`` 之类的字符，则它将单独保留。 ``as_needed`` 不支持此选项。
        默认为 ``True`` 。

    Returns
    --------
    :class:`str`
        带有 Markdown 特殊字符的文本用斜杠转义。
    """

    if not as_needed:

        def replacement(match):
            groupdict = match.groupdict()
            is_url = groupdict.get('url')
            if is_url:
                return is_url
            return '\\' + groupdict['markdown']

        regex = _MARKDOWN_STOCK_REGEX
        if ignore_links:
            regex = f'(?:{_URL_REGEX}|{regex})'
        return re.sub(regex, replacement, text, 0, re.MULTILINE)
    else:
        text = re.sub(r'\\', r'\\\\', text)
        return _MARKDOWN_ESCAPE_REGEX.sub(r'\\\1', text)


class _cached_property:
    def __init__(self, function):
        self.function = function
        self.__doc__ = getattr(function, '__doc__')

    def __get__(self, instance, owner):
        if instance is None:
            return self

        value = self.function(instance)
        setattr(instance, self.function.__name__, value)

        return value


if TYPE_CHECKING:
    from functools import cached_property as cached_property
    from typing_extensions import ParamSpec

    P = ParamSpec('P')

else:
    cached_property = _cached_property

try:
    import orjson
except ModuleNotFoundError:
    HAS_ORJSON = False
else:
    HAS_ORJSON = True

if HAS_ORJSON:

    def _to_json(obj: Any) -> str:  # type: ignore
        return orjson.dumps(obj).decode('utf-8')


    _from_json = orjson.loads  # type: ignore

else:

    def _to_json(obj: Any) -> str:
        return json.dumps(obj, separators=(',', ':'), ensure_ascii=True)


    _from_json = json.loads


class _MissingSentinel:
    def __eq__(self, other):
        return False

    def __bool__(self):
        return False

    def __repr__(self):
        return '...'


MISSING: Any = _MissingSentinel()


async def maybe_coroutine(f, *args, **kwargs):
    value = f(*args, **kwargs)
    if _isawaitable(value):
        return await value
    else:
        return value


def copy_doc(original: Callable) -> Callable[[T], T]:
    def decorator(overriden: T) -> T:
        overriden.__doc__ = original.__doc__
        overriden.__signature__ = _signature(original)  # type: ignore
        return overriden

    return decorator


@overload
def parse_time(timestamp: None) -> None:
    ...


@overload
def parse_time(timestamp: str) -> datetime.datetime:
    ...


@overload
def parse_time(timestamp: Optional[str]) -> Optional[datetime.datetime]:
    ...


def parse_time(timestamp: Optional[str]) -> Optional[datetime.datetime]:
    if timestamp:
        if timestamp.isnumeric():
            return datetime.datetime.fromtimestamp(int(timestamp))
        return datetime.datetime.fromisoformat(timestamp)
    return None


def _unique(iterable: Iterable[T]) -> List[T]:
    return [x for x in dict.fromkeys(iterable)]


class SnowflakeList(array.array):
    __slots__ = ()

    if TYPE_CHECKING:
        def __init__(self, data: Iterable[int], *, is_sorted: bool = False):
            ...

    def __new__(cls, data: Iterable[int], *, is_sorted: bool = False):
        return array.array.__new__(cls, 'Q', data if is_sorted else sorted(data))  # type: ignore

    def add(self, element: int) -> None:
        i = bisect_left(self, element)
        self.insert(i, element)

    def get(self, element: int) -> Optional[int]:
        i = bisect_left(self, element)
        return self[i] if i != len(self) and self[i] == element else None

    def has(self, element: int) -> bool:
        i = bisect_left(self, element)
        return i != len(self) and self[i] == element


class CachedSlotProperty(Generic[T, T_co]):
    def __init__(self, name: str, function: Callable[[T], T_co]) -> None:
        self.name = name
        self.function = function
        self.__doc__ = getattr(function, '__doc__')

    @overload
    def __get__(self, instance: None, owner: Type[T]):
        ...

    @overload
    def __get__(self, instance: T, owner: Type[T]) -> T_co:
        ...

    def __get__(self, instance: Optional[T], owner: Type[T]) -> Any:
        if instance is None:
            return self

        try:
            return getattr(instance, self.name)
        except AttributeError:
            value = self.function(instance)
            setattr(instance, self.name, value)
            return value


def cached_slot_property(name: str) -> Callable[[Callable[[T], T_co]], CachedSlotProperty[T, T_co]]:
    def decorator(func: Callable[[T], T_co]) -> CachedSlotProperty[T, T_co]:
        return CachedSlotProperty(name, func)

    return decorator


def escape_mentions(text: str) -> str:
    """一个帮助函数，可以转义所有成员，身份组和用户提及。

    .. note::

        这不包括频道提及。

    .. note::

        要对消息中的提及内容进行更精细的控制，请参阅 :class:`~qq.AllowedMentions` 类。

    Parameters
    -----------
    text: :class:`str`
        要转义的文本。

    Returns
    --------
    :class:`str`
        删除了提及的文本。
    """

    return re.sub(r'@(所有成员|[!&]?[0-9]{17,20})', '@\u200b\\1', text)


def find(predicate: Callable[[T], Any], seq: Iterable[T]) -> Optional[T]:
    """返回在满足 predicate 的序列中找到的第一个元素的帮助器。例如： ::

        member = qq.utils.find(lambda m: m.name == 'Foo', channel.guild.members)

    会找到第一个名字是 ``Mighty`` 的  :class:`~qq.Member` 并返回它。
    如果未找到条目，则返回 ``None`` 。
    这与 :func:`py:filter` 不同，因为它在找到有效条目时停止。

    Parameters
    -----------
    predicate
        返回类似布尔值的结果的函数。
    seq: :class:`collections.abc.Iterable`
        要搜索的可迭代对象。
    """

    for element in seq:
        if predicate(element):
            return element
    return None


def get(iterable: Iterable[T], **attrs: Any) -> Optional[T]:
    r"""一个帮助器，它返回可迭代对象中满足 ``attrs`` 中传递的所有特征的第一个元素。这是 :func:`~qq.utils.find` 的替代方案。
    指定多个属性时，将使用逻辑 AND 而不是逻辑 OR 检查它们。
    这意味着他们必须满足传入的每个属性，而不是其中之一。
    要进行嵌套属性搜索（即通过 ``x.y`` 搜索），然后传入 ``x__y`` 作为关键字参数。
    如果没有找到与传递的属性匹配的属性，则返回 ``None`` 。

    Examples
    ---------

    基本用法:

    .. code-block:: python3

        member = qq.utils.get(message.guild.members, name='Foo')

    多属性匹配:

    .. code-block:: python3

        member = qq.utils.get(message.guild.members, name='Foo', bot=False)

    嵌套属性匹配：

    .. code-block:: python3

        member = qq.utils.get(message.guild.members, avatar__url='xxx', name='Foo')

    Parameters
    -----------
    iterable
        一个可迭代的搜索对象。
    \*\*attrs
        表示要搜索的属性的关键字参数。
    """

    _all = all
    attrget = attrgetter

    # Special case the single element call
    if len(attrs) == 1:
        k, v = attrs.popitem()
        pred = attrget(k.replace('__', '.'))
        for elem in iterable:
            if pred(elem) == v:
                return elem
        return None

    converted = [(attrget(attr.replace('__', '.')), value) for attr, value in attrs.items()]

    for elem in iterable:
        if _all(pred(elem) == value for pred, value in converted):
            return elem
    return None


async def sane_wait_for(futures, *, timeout):
    ensured = [asyncio.ensure_future(fut) for fut in futures]
    done, pending = await asyncio.wait(ensured, timeout=timeout, return_when=asyncio.ALL_COMPLETED)

    if len(pending) != 0:
        raise asyncio.TimeoutError()

    return done


def compute_timedelta(dt: datetime.datetime):
    if dt.tzinfo is None:
        dt = dt.astimezone()
    now = datetime.datetime.now(datetime.timezone.utc)
    return max((dt - now).total_seconds(), 0)


PY_310 = sys.version_info >= (3, 10)


def flatten_literal_params(parameters: Iterable[Any]) -> Tuple[Any, ...]:
    params = []
    literal_cls = type(Literal[0])
    for p in parameters:
        if isinstance(p, literal_cls):
            params.extend(p.__args__)
        else:
            params.append(p)
    return tuple(params)


def normalise_optional_params(parameters: Iterable[Any]) -> Tuple[Any, ...]:
    none_cls = type(None)
    return tuple(p for p in parameters if p is not none_cls) + (none_cls,)


def evaluate_annotation(
        tp: Any,
        globals: Dict[str, Any],
        locals: Dict[str, Any],
        cache: Dict[str, Any],
        *,
        implicit_str: bool = True,
):
    if isinstance(tp, ForwardRef):
        tp = tp.__forward_arg__
        # ForwardRefs always evaluate their internals
        implicit_str = True

    if implicit_str and isinstance(tp, str):
        if tp in cache:
            return cache[tp]
        evaluated = eval(tp, globals, locals)
        cache[tp] = evaluated
        return evaluate_annotation(evaluated, globals, locals, cache)

    if hasattr(tp, '__args__'):
        implicit_str = True
        is_literal = False
        args = tp.__args__
        if not hasattr(tp, '__origin__'):
            if PY_310 and tp.__class__ is types.UnionType:  # type: ignore
                converted = Union[args]  # type: ignore
                return evaluate_annotation(converted, globals, locals, cache)

            return tp
        if tp.__origin__ is Union:
            try:
                if args.index(type(None)) != len(args) - 1:
                    args = normalise_optional_params(tp.__args__)
            except ValueError:
                pass
        if tp.__origin__ is Literal:
            if not PY_310:
                args = flatten_literal_params(tp.__args__)
            implicit_str = False
            is_literal = True

        evaluated_args = tuple(
            evaluate_annotation(arg, globals, locals, cache, implicit_str=implicit_str) for arg in args)

        if is_literal and not all(isinstance(x, (str, int, bool, type(None))) for x in evaluated_args):
            raise TypeError('Literal arguments must be of type str, int, bool, or NoneType.')

        if evaluated_args == args:
            return tp

        try:
            return tp.copy_with(evaluated_args)
        except AttributeError:
            return tp.__origin__[evaluated_args]

    return tp


def resolve_annotation(
        annotation: Any,
        globalns: Dict[str, Any],
        localns: Optional[Dict[str, Any]],
        cache: Optional[Dict[str, Any]],
) -> Any:
    if annotation is None:
        return type(None)
    if isinstance(annotation, str):
        annotation = ForwardRef(annotation)

    locals = globalns if localns is None else localns
    if cache is None:
        cache = {}
    return evaluate_annotation(annotation, globalns, locals, cache)



#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING

from . import abc
from .enum import AudioStatusType, try_enum

if TYPE_CHECKING:
    from .types.audio import (
        StartAudioControl as StartAudioControlPayload,
        PauseAudioControl as PauseAudioControlPayload,
        ResumeAudioControl as ResumeAudioControlPayload,
        StopAudioControl as StopAudioControlPayload,
        AudioAction as AudioActionPayload
    )


class AudioAction:
    """代表音频频道中的状态

    .. container:: operations
        .. describe:: str(x)

            返回状态文本。

    Attributes
    -----------
    audio_url: Optional[:class:`str`]
        音频数据的url，可能为 ``None`` 。
    text: Optional[:class:`Guild`]
        状态文本，可能为 ``None`` 。
    guild_id: :class:`int`
        音频的频道 ID。
    channel_id: Optional[:class:`int`]
        音频的子频道 ID。
    """

    __slots__ = (
        'audio_url',
        'text',
        'guild_id',
        'channel_id'
    )

    def __init__(self, data: AudioActionPayload):
        self.audio_url = data.get('audio_url')
        self.text = data.get('text')
        self.guild_id = data.get('guild_id')
        self.channel_id = data.get('channel_id')


class StartAudio(abc.BaseAudioControl):
    """代表音频频道中的开始播放动作

    .. container:: operations
        .. describe:: str(x)

            返回状态文本。

    """

    __slots__ = (
        '_audio_url',
        '_text',
        '_status'
    )

    def __init__(self, data: StartAudioControlPayload):
        self._status = data.get('status', '')
        self._text = data.get("text", '')
        self._audio_url = data.get("audio_url", '')

    @property
    def type(self) -> AudioStatusType:
        """返回动作的种类
        """
        return try_enum(AudioStatusType, self._status)

    def to_dict(self):
        return {"audio_url": self._audio_url, "text": self._text, "status": self.type}


class PauseAudio(abc.BaseAudioControl):
    """代表音频频道中的暂停播放动作

    .. container:: operations
        .. describe:: str(x)

            返回状态文本。

    """

    __slots__ = (
        '_audio_url',
        '_text',
        '_status'
    )

    def __init__(self, data: PauseAudioControlPayload):
        self._status = data.get('status', '')
        self._text = data.get("text", '')
        self._audio_url = data.get("audio_url", '')

    @property
    def type(self) -> AudioStatusType:
        """返回动作的种类
        """
        return try_enum(AudioStatusType, self._status)

    def to_dict(self):
        return {"status": self.type}


class ResumeAudio(abc.BaseAudioControl):
    """代表音频频道中的继续播放动作

    .. container:: operations
        .. describe:: str(x)

            返回状态文本。

    """

    __slots__ = (
        '_audio_url',
        '_text',
        '_status'
    )

    def __init__(self, data: ResumeAudioControlPayload):
        self._status = data.get('status', '')
        self._text = data.get("text", '')
        self._audio_url = data.get("audio_url", '')

    @property
    def type(self) -> AudioStatusType:
        """返回动作的种类
        """
        return try_enum(AudioStatusType, self._status)

    def to_dict(self):
        return {"status": self.type}


class StopAudio(abc.BaseAudioControl):
    """代表音频频道中的停止播放动作

    .. container:: operations
        .. describe:: str(x)

            返回状态文本。

    """

    __slots__ = (
        '_audio_url',
        '_text',
        '_status'
    )

    def __init__(self, data: StopAudioControlPayload):
        self._status = data.get('status', '')
        self._text = data.get("text", '')
        self._audio_url = data.get("audio_url", '')

    @property
    def type(self) -> AudioStatusType:
        """返回动作的种类
        """
        return try_enum(AudioStatusType, self._status)

    def to_dict(self):
        return {"status": self.type}



__path__ = __import__('pkgutil').extend_path(__path__, __name__)
__title__ = 'qq'
__author__ = 'Foxwhite'
__license__ = 'MIT'
__version__ = '1.0.13'

#  The MIT License (MIT)
#  Copyright (c) 2021-present foxwhite25
#
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.

import logging
from typing import NamedTuple, Literal
from .client import *
from .user import *
from .channel import *
from .guild import *
from .flags import *
from .member import *
from .message import *
from .asset import *
from .role import *
from .file import *
from .colour import *
from .object import *
from . import utils, abc
from .enum import *
from .shard import *
from .error import *
from .embeds import *
from .mention import *
from .partial_emoji import *
from .raw_models import *
from .permissions import *


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: Literal["alpha", "beta", "candidate", "final"]
    serial: int


version_info: VersionInfo = VersionInfo(major=1, minor=0, micro=13, releaselevel='beta', serial=0)
